"""PyECDAR, a open source software for timed systems.
Copyright (C) 2013 INRIA Rennes

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA."""

import xml.etree.cElementTree as ElementTree
import logging
import re
import itertools

import pydbm
import pycudd
import pysymbolicstates

from misc_functions import evaluate_expression, split_constraints

logging.basicConfig()
logger = logging.getLogger('PYNTA')


""" Structure for networks of timed i/o automata:
- class NTA: network of timed i/o automata
- class Template: single timed i/o automata
- class Location: location in a timed i/o automata
- class Transition: transition in a timed i/o automata
- class Variable: fixed range integer variable
"""

#TODO: add support for arrays
#TODO: add parameters to templates
#TODO: add support for select fields in transitions
    
class NTA:
  """ A network of timed i/o automata composed by:
  - a list of communication channels
  - a list of templates
  - a list of constant variables with their value
  """
  
  def __init__(self):
    """ Build an empty NTA"""
    self.channels = []		# list of channels name
    self.templates = []		# list of templates
    self.constants = {}  	# dictionary with constant name and value

  def add_template(self, t):
    if not t in self.templates:
      self.templates.append(t)
      
  def get_template(self,name):
    ts = [t for t in self.templates if t.name == name] 
    if len(ts) == 0:
      logger.error("No template with name \""+name+"\"")
      return None
    elif len(ts) > 1:
      logger.error("More than one template with name \""+name+"\"")
      return None
    else:
      return ts[0]
  
  @classmethod
  def from_xml(cls, xmlsock):
    """Static method to load a model from an xml file"""
    nta = cls()
    nta._from_xml(xmlsock)
    return nta
  
  def _from_xml(self, xmlsock):
    xmldoc = ElementTree.ElementTree(file=xmlsock).getroot()
    
    declaration = xmldoc.findtext("declaration") or ""
    self.parse_declaration(declaration)
      
    self.templates = []
    for templatexml in xmldoc.getiterator("template"):      
      template = Template(self,templatexml.find("name").text)
      locations = {}
      for locationxml in templatexml.getiterator("location"):
	name = locationxml.findtext("name") or ""
	id = locationxml.get('id')
        location = Location(template,name)
	if locationxml.find("urgent") != None:
	  location.urgent = True
	for labelxml in locationxml.getiterator("label"):
	  if labelxml.get('kind') == 'invariant':
	    location.invariant = labelxml.text.replace("\n"," ")
	locations[id] = location
      transitions = []
      for transitionxml in templatexml.getiterator("transition"):
	transition = Transition(locations[transitionxml.find('source').get('ref')],locations[transitionxml.find('target').get('ref')])
	transition.source.add_forward_transition(transition)
	transition.target.add_backward_transition(transition)
	transition.controllable = ('controllable', 'false') not in transitionxml.items() # parsed but changed according to synchronisation
        for labelxml in transitionxml.getiterator("label"):
	  if labelxml.get('kind') == 'assignment':
	    transition.update = labelxml.text.replace("\n"," ")
	  elif labelxml.get('kind') == 'guard':
	    transition.guard = labelxml.text.replace("\n"," ")  
	  elif labelxml.get('kind')  == 'synchronisation':
	    sync_type = labelxml.text[-1]
	    if sync_type == '?':
	      transition.controllable = False 	# this is not like in ECDAR
	    elif sync_type == '!':
	      transition.controllable = True
	    else:
	      logger.error("Unknown synchronisation type: \'"+sync_type+"\'")
	    channel_name = labelxml.text[:-1]
	    if channel_name in self.channels:
	      transition.synchronisation = (channel_name,sync_type)
	transitions.append(transition)
	  
      if templatexml.find("init") != None:
	init_location = locations[templatexml.find("init").get('ref')]
      else:
        init_location = None
        logger.warning("The template \'" + template.name + "\' has no initial location.")
      template.locations = locations.values()
      template.transitions = transitions
      template.init_location = init_location
      self.templates.append(template)
      template.parse_declaration(templatexml.findtext("declaration") or "")
    if not self.templates:
      logger.warning("The model has no template.")
    
  def to_xml(self):
    ret = """<?xml version="1.0" encoding="utf-8"?>
    <!DOCTYPE nta PUBLIC "-//Uppaal Team//DTD Flat System 1.1//EN" "http://www.it.uu.se/research/group/darts/uppaal/flat-1_1.dtd">
    <nta>
      <declaration>"""
    for c in self.channels:
      ret += "broadcast chan " + c + ";\n"
    for c in self.constants:
      if isinstance(c,bool):
	ret += "const bool " + c + " = "
	if self.constants[c]:
	  ret += "true;\n"
	else:
	  ret += "false;\n"
      else:
	ret += "const int " + c + " = " + str(self.constants[c]) + ";\n"
    ret += "</declaration>\n"
    
    for t in self.templates:
      ret += t.to_xml() + "\n"
    
    if self.templates:
      ret += "<system>"
      ret += self.templates[0].name
      for t in self.templates[1:]:
	ret += ", " + t.name
      ret += "</system></nta>"
    return ret    
    
  def parse_declaration(self,declaration_string):
    if declaration_string:
      decls = declaration_string.split('\n')   
      for decl in decls:
	if decl:
	  r = re.search('^\s*//.*',decl)
	  if r:
	    continue
	  r = re.search('^\s*broadcast\s*chan\s*(.*);',decl)
	  if r:
	    chan_names = r.group(1)
	    chan_names = chan_names.split(',')
	    for c in chan_names:
	      self.channels.append(c.strip())
	    continue
	  r = re.search('^\s*const\s*(bool|int)\s*(\w+)\s*=(.*);',decl)
	  if r:
	    val = evaluate_expression(r.group(3).strip(),self.constants)
	    assert(isinstance(val, int) or isinstance(val, bool))
	    self.constants[r.group(2)] = val
	    continue
	  logger.warning('Cannot parse declaration "' + decl + '"')
	  

class Template:
  """ Structure of a timed i/o automata:
  - a list of of locations
  - a list of transitions
  - an initial location
  - a list of real time clocks linked to the clocks from Uppaal DBM Libary
  - a list of local variables linked to their global name (prefix by the name of the template)
  - the list of all the global variable names that may be used in the template (= local variables + nta variables)
  - lists of features
  - a DBM context
  - integer bounds for real time clocks
  - a BDD manager 
  """
  
  def __init__(self,nta,name,locations=None,transitions=None,init_location=None,clocks=None,constants=None,variables=None,fs=None,fa=None,fb=None,fe=None,ff=None):
    """ Build an empty template or initialised with a set of elements"""
    self.nta = nta
    self.name = name
    self.locations = locations or []
    self.transitions = transitions or [] 
    self.init_location = init_location
    
    self.clocks = clocks or {} 			# dictionary: links a local clock name to the real time clock in the DBM context
    self.constants = constants or {} 		# dictionary with constant name and value
    self.variables = variables or {} 		# dictionary with a variable name and variable

    self.fs = fs or [] # system feature
    self.fa = fa or [] # adaptive features
    self.fb = fb or [] # new adaptive features (same order as fa)    
    self.fe = fe or [] # environment feature
    self.ff = ff or [] # new environment feature (same order as fe)
    self.features = dict([(k,None) for k in self.fs + self.fa + self.fb + self.fe + self.ff]) # dictionary of features: links a feature name to a CUDD variable. Updated after template parsing.
    
    self.context = None 	# context from the Uppaal DBM library. Updated after template parsing.
    self.clock_bounds = {}	# dictionary: links a real time clock from the DBM context to an integer maximum bound. Updated after template parsing.
    self.mgr = None		# manager for BDD from pyCUDD. Updated after template parsing.
    
  def get_location(self,loc_name):
    ls = [l for l in self.locations if l.name == loc_name] 
    if len(ls) == 0:
      return None
    elif len(ls) > 1:
      logger.warning("More than one location with name \""+loc_name+"\"")
      return None
    else:
      return ls[0]
  
  def get_initial_discrete_state(self):
    """The initial state is composed by the initial location and a tuple of initial values for the integer variables"""
    return pysymbolicstates.DiscreteState(self.init_location,tuple(v.init_val for v in self.variables.values()))
  
  def get_list_discrete_states(self):
    """Return the list of all the possible discrete states in the template: this is the cartesian product of all the locations and all the values of the integer variables"""
    list_values = [v.get_values() for v in self.variables.values()]
    list_values = [t for t in itertools.product(*list_values)]
    list_states = []
    for l in self.locations:
      if list_values:
	for v in list_values:
	  list_states.append(pysymbolicstates.DiscreteState(l,v))
      else:
	list_states.append(pysymbolicstates.DiscreteState(l,None))
    return list_states
  
  def get_nb_features(self):
    return len(self.fa) + len(self.fe) + len(self.fs) 
  
  def get_nb_adaptive_features(self):
    return len(self.fa) + len(self.fe)
    
  def to_xml(self):
    ret = "<template><name>" + self.name + "</name>"
    ret += "<declaration>"
    for c in self.constants:
      if isinstance(c,bool):
	ret += "const bool " + c + " = "
	if self.constants[c]:
	  ret += "true;\n"
	else:
	  ret += "false;\n"
      else:
	ret += "const int " + c + " = " + self.constants[c] + ";\n"
    for var in self.variables:
      ret += "int[" + var.min_val + "," + var.max_val + "] " + var.name + " = " + var.init_val + ";\n"
    for c in self.clocks:
      ret += "clock " + c + ";\n"
    if self.fs:
      ret +=  "// system features\n"
      for f in self.fs:
	ret += "bool " + f + ";\n"
    if self.fa:
      ret +   "// adaptive features\n"
      for index in range(0,len(self.fa)):
	ret += "bool " + self.fa[index] + ", " + self.fb[index] + ";\n"
    if self.fe:
      ret +=  "// environment features\n"
      for index in range(0,len(self.fe)):
	ret += "bool " + self.fe[index] + ", " + self.ff[index] + ";\n"
    ret += "</declaration>\n"
    
    locations_id = {}
    lid = 0
    for l in self.locations:
      ret += l.to_xml(lid)
      ret += "\n"
      locations_id[l] = lid
      lid += 1
    if self.init_location:
      ret += "<init ref=\"id" + str(locations_id[self.init_location]) + "\"/>"  
    for t in self.transitions:
      ret += t.to_xml(locations_id)
      ret += "\n"
    ret += "</template>"  
    return ret
    
  def updateBound(self,clock, value):
    if self.clock_bounds[clock] is None:
      self.clock_bounds[clock] = value
    else:
      self.clock_bounds[clock] = max(self.clock_bounds[clock], value)
    
  def parse_declaration(self,declaration_string):
    if declaration_string:
      decls = declaration_string.split('\n')
      parse_feature = ""
      for decl in decls:
	decl = decl.strip()
	if not decl:
	  continue
	if decl == "// system features":
	  parse_feature = "fs"
	  continue
	if decl == "// adaptive features":
	  parse_feature = "fa"
	  continue
	if decl == "// environment features":
	  parse_feature = "fe"
	  continue
	
	if parse_feature:
	  if decl.startswith("bool "):
	    if parse_feature == "fa":
	      r = re.search('^bool\s+(\w+)\s*,\s*(\w+).*$', decl)
	      if r:
		self.fa.append(r.group(1))
		self.fb.append(r.group(2))
		self.features[r.group(1)] = None
		self.features[r.group(2)] = None
		continue
	    if parse_feature == "fs":
	      r = re.search('^bool\s+(\w+)\s*.*$', decl)
	      if r:
		self.fs.append(r.group(1))
		self.features[r.group(1)] = None
		continue
	    if parse_feature == "fe":
	      r = re.search('^bool\s+(\w+)\s*,\s*(\w+).*$', decl)
	      if r:
		self.fe.append(r.group(1))
		self.ff.append(r.group(2))
		self.features[r.group(1)] = None
		self.features[r.group(2)] = None
		continue
	  logger.warning("Format not supported for feature: "+decl)
	  continue
	
	r = re.search('^\s*//.*',decl)
	if r:
	  continue
	r = re.search('^\s*clock\s*(.*);',decl)
	if r:
	  clock_names = r.group(1)
	  clock_names = clock_names.split(',')
	  for c in clock_names:
	    self.clocks[c.strip()] = None
	  #WARNING: no syntax check is performed. clocks may be declared two times, etc...
	  # we assume the model is checked in UPPAAL/ECDAR
	  continue
	r = re.search('^\s*const\s*(bool|int)\s*(\w+)\s*=(.*);',decl)
	if r:
	  val = evaluate_expression(r.group(3),dict(self.constants.items() + self.nta.constants.items()))
	  assert(isinstance(val, int) or isinstance(val, bool))
	  self.constants[r.group(2)] = val
	  continue
	r = re.search('^\s*bool\s*(\w+)',decl)
	if r:
	  v = Variable(r.group(1),False,0,1) 		# default initial value is False
	  self.variables[v.name] = v;
	  continue
	r = re.search('^\s*int\s*\[(.*),(.*)\]\s*(\w+)',decl)
	if r:
	  min_val = evaluate_expression(r.group(1),dict(self.constants.items() + self.nta.constants.items()))
	  max_val = evaluate_expression(r.group(2),dict(self.constants.items() + self.nta.constants.items()))
	  v = Variable(r.group(3),min_val,min_val,max_val) 	# default initial value is min_val
	  self.variables[v.name] = v;
	  continue
	r = re.search('^\s*bool\s*(\w+)\s*=(.*)',decl)
	if r:
	  val = evaluate_expression(r.group(2),dict(self.constants.items() + self.nta.constants.items()))
	  assert(isinstance(val,bool))
	  v = Variable(r.group(1),val,0,1)
	  self.variables[v.name] = v;
	  continue
	r = re.search('^\s*int\s*\[(.*),(.*)\]\s*(\w+)\s*=(.*)',decl)
	if r:
	  min_val = evaluate_expression(r.group(1),dict(self.constants.items() + self.nta.constants.items()))
	  max_val = evaluate_expression(r.group(2),dict(self.constants.items() + self.nta.constants.items()))
	  val = evaluate_expression(r.group(4),dict(self.constants.items() + self.nta.constants.items()))
	  assert(val <= max_val and val >= min_val)
	  v = Variable(r.group(3),val,min_val,max_val)
	  self.variables[v.name] = v;
	  continue
	logger.warning('Cannot parse declaration "' + decl + '"')

  def parse_template(self,tadactl=False):
    #TODO: bug. reparsing seems sometimes provoques a segmentation fault. Problems with CUDD
    """ Creates a DBM context with all the clocks plus optionaly a special tadactl_clock
	If the template has features, creates a manager.
	Parse the guards and the invariants to build the DBMs and the BDDs.
    """ 
    self.context = pydbm.Context()
    self.clock_bounds = {}
    self.mgr = None
    for clock_name in self.clocks:
      self.context.addClockByName(clock_name)
      clock = self.context[clock_name]
      self.clocks[clock_name] = clock
      self.clock_bounds[clock] = None
    if tadactl:
      self.context.addClockByName("tadactl_clock")
      clock = self.context["tadactl_clock"]
      self.clocks["tadactl_clock"] = clock
      self.clock_bounds[clock] = None

    # create a BDD manager and add the features variables
    if len(self.features) > 0:
      self.mgr = pycudd.DdManager(len(self.features))
      self.mgr.SetDefault()
      pysymbolicstates.SetOfSymbolicStates.mgr = self.mgr
      index = 0
      for feature in self.fs:
	self.features[feature] = self.mgr.IthVar(index)
	index += 1
      for feature in self.fa:
	self.features[feature] = self.mgr.IthVar(index)
	index += 1
      for feature in self.fe:
	self.features[feature] = self.mgr.IthVar(index)
	index += 1
      for feature in self.fb:
	self.features[feature] = self.mgr.IthVar(index)
	index += 1
      for feature in self.ff:
	self.features[feature] = self.mgr.IthVar(index)
	index += 1
    
    # update the variables id to fit the order of the variables
    i = 0
    for var in self.variables.values():
      var.id = i
      i += 1

    #parse the guards and invariants into DBMs and BDDs
    for location in self.locations:
      location.parse_invariant()
    for transition in self.transitions:
      transition.parse_guard()
      transition.parse_update()
    
  def parse_dbm(self, constraints):
    """ Parse a constraint string into a DBM.
	Return None if errors occur.
    """
    #print "parse dbm: " + constraints + "\n"
    if not constraints:
      return None
    constraints = constraints.replace('&&','and')
    constraints = constraints.replace('|','or')
    dbm = evaluate_expression(constraints,dict(self.clocks.items() + self.nta.constants.items() + self.constants.items()))
    if isinstance(dbm, pydbm.Federation):
      #print str(dbm)
      constraints_dbm = str(dbm)[1:-1]
      constraints_dbm = constraints_dbm.split('&')
      for c in constraints_dbm:
	c = c.strip()
	if c:
	  r = re.search('^(?P<clock>.+)(<=|<|==)(?P<coeff>\d+)$',c)
	  if r is None:
	    r = re.search('^(?P<coeff>\d+)(<=|<)(?P<clock>.+)$',c)
	  if r is None:
	    logger.error('Unexpected error in dbm constraint ' + c)
	    continue
	  clock = self.context[r.group("clock")]
	  value = int(r.group('coeff'))
	  self.updateBound(clock,value)
    else:
      dbm = None
    return dbm
    
  def parse_bdd(self, expr):
    """ Parse a feature expression into a BDD with the manager of the template
	Return None is errors
    """
    #print "parse bdd: " + expr + "\n"
    if not expr or self.get_nb_features() == 0:
      return None
    expr = expr.strip()  
    expr = expr.replace('&&','&')
    expr = expr.replace('and','&')
    expr = expr.replace('||','|')
    expr = expr.replace('or','|')
    expr = expr.replace('!','~')
    expr = expr.replace('not','~')
    expr = expr.replace('true','True')
    expr = expr.replace('false','False')
    bdd = evaluate_expression(expr,self.features)
    return bdd

  #def parse_variables_expression(self, expr):
    #""" Repace in expr all the constants by their value"""
    #new_expr = ""
    #while expr:
      #r = re.search('^([^\w]*)(\w*)(.*)$',expr)
      #assert(r)
      #start = r.group(1) or ""
      #identifier = r.group(2)
      #end = r.group(3)
      #if end == expr:
	#new_expr +=  end
	#expr = ""
      #else:
	#if self.constants.has_key(identifier):
	  #new_expr += start + str(self.constants[identifier])
	  #expr = end
	#elif self.nta.constants.has_key(identifier):
	  #new_expr += start + str(self.nta.constants[identifier])
	  #expr = end
	#else:
	  #new_expr += start + identifier
	  #expr = end
    #return new_expr

  def bdd_to_str(self,bdd):
    """Print a BDD into a string such that variable are replaced by the name of the feature"""
    if bdd.IsConstant():
      if bdd.Empty():
	return "false"
      else:
	return "true"
    ret=""
    pycudd.set_iter_meth(0)
    for prime in bdd:
      if ret:
	ret += " or "
      ret += "("
      index = 0
      printed = False
      for p in prime:
	if p != 2:
	  if printed:
	    ret += " and "
	  if p == 0:
	    ret += "not" + self.get_feature_name(index)
	  elif p == 1:
	    ret += self.get_feature_name(index)
	  printed = 1
  	index +=1
      ret += ")"
    return ret

  def get_feature_name(self,index):
    """Return the name of the feature according to its index.
    The order that we assume is: fs - fa - fe - fb - ff""" 
    if index < len(self.fs):
      return self.fs[index]
    index -= len(self.fs)
    if index < len(self.fa):
      return self.fa[index]
    index -= len(self.fa)
    if index < len(self.fe):
      return self.fe[index]
    index -= len(self.fe)
    if index < len(self.fb):
      return self.fb[index]
    index -= len(self.fb)
    if index < len(self.fe):
      return self.ff[index]
    return ""
  
  def update_forward_backward(self):
    for l in self.locations:
      l.forward = []
      l.backward = []
    for t in self.transitions:
      t.source.add_forward_transition(t)
      t.target.add_backward_transition(t)      
      
class Location:
  """ A location in a TIOA."""
  
  def __init__(self,template,name,urgent=False,invariant=""):
    self.template = template
    self.name = name
    self.urgent = urgent
    self.invariant = invariant		# original invariant in a string
    self.invariant_dbm = None		# clock invariant in a DBM	
    self.invariant_bdd = None		# feature invariant in a BDD
    self.invariant_var = None		# variable invariant in a string
    self.forward = []			# list of outgoing transition from this location
    self.backward = []			# list of incomming transition to this location
        
  def add_forward_transition(self,t):
    self.forward.append(t)
        
  def add_backward_transition(self,t):
    self.backward.append(t)
    
  def parse_invariant(self):
    """ Split the invariant string into three invariants (for clocks, features, and variables) and creates their respective domain (DBM or BDD)"""
    self.invariant_dbm = self.template.context.getTautologyFederation()
    self.invariant_bdd = None
    self.invariant_var = ""
    constraints = split_constraints(self.invariant)
    if len(self.template.features) > 0:
      self.invariant_bdd = self.template.mgr.One()
    for cons in constraints:
      cons = cons.strip()
      if cons:
	dbm = self.template.parse_dbm(cons)
	if dbm:
	  self.invariant_dbm &= dbm
	  continue
	if self.invariant_bdd:
	  bdd = self.template.parse_bdd(cons)
	  if not bdd is None:
	    self.invariant_bdd &= bdd
	    continue
	if self.invariant_var:
	  self.invariant_var += " and " + cons
	else:
	  self.invariant_var = cons
	#logger.error("Cannot parse contraint \"" + cons + "\" in invariant of location \"" + self.name + "\"")
    if not self.invariant_dbm.hasZero():
      logger.error("Invariant of location \"" + str(self.name) + "\" is not an upper bound")   

  def to_xml(self,lid):
    ret = "<location id=\"id" + str(lid) + "\">"
    if self.name:
      ret += "<name>" + self.name + "</name>"
    
    #invariant_str = str(self.invariant_dbm)
    #if invariant_str != "true":
      #invariant_str = invariant_str.replace("&","&amp;")
      #invariant_str = invariant_str.replace("<","&lt;")
    #else:
      #invariant_str = ""
    #if self.invariant_bdd:
      #invariant_bdd_str = self.template.bdd_to_str(self.invariant_bdd)
      #if invariant_bdd_str != "true":
	#invariant_bdd_str = invariant_bdd_str.replace("~","not ")
	#if invariant_str:
	  #invariant_str += " and (" + invariant_bdd_str + ")"
	#else:
	  #invariant_str = "(" + invariant_bdd_str + ")"
    #if self.invariant_var:
      #if invariant_str:
	#invariant_str += " and (" + self.invariant_var + ")"
      #else:
	#invariant_str = "(" + self.invariant_var + ")"
	
    if self.invariant:
      invariant_xml = self.invariant
      invariant_xml = invariant_xml.replace("&","&amp;")
      invariant_xml = invariant_xml.replace("<","&lt;")
      invariant_xml = invariant_xml.replace(">","&gt;")
      ret += "<label kind=\"invariant\">" + invariant_xml + "</label>"
    if self.urgent:
      ret += "<urgent/>"
    ret += "</location>"
    return ret
  

class Transition:
  """ A transition in a TIOA"""
  
  def __init__(self,source,target,select=None,guard="",synchronisation=None,update="",controllable=True):
    assert(source.template == target.template)
    self.template = source.template
    self.source = source	# source location of the transition
    self.target = target	# target location of the transition
    self.select = select	# select fields (not used in current version)
    self.guard = guard		# original guard in a string
    self.guard_dbm = None	# clock guard in a DBM
    self.guard_bdd = None	# features guard in a BDD
    self.guard_var = None	# variables guard in a string
    self.synchronisation = synchronisation	# synchronisation of the transition: tuple with a channel name and the modality (either "?" for inputs or "!" for outputs)
    self.update = update	# original assignments in a string
    self.update_clock = []	# list of clocks reset by the transition
    self.update_var = []	# list of variables updates: each one is a tuple with a variable, a "+" (if increment) or "-" (if decrement) sign, and an assignment expression in a string
    self.controllable = controllable	# controllability of the transition: for timed specifications True if output, False if input (not like Ecdar)

  def to_xml(self,locations_id):
    ret = "<transition"
    if not self.controllable:
      ret += " controllable=\"false\""
    ret += ">"  
    ret += "<source ref=\"id" + str(locations_id[self.source]) + "\"/>"
    ret += "<target ref=\"id" + str(locations_id[self.target]) + "\"/>"
    
    #if self.guard_dbm.getSize() > 1:
      #logger.error("Guard of transition from \"" + str(self.source.name) + "\" to \""  + str(self.target.name) + "\" is a federation of DBMs. It cannot be printed to XML.")
      #return ""
    #guard_str = str(self.guard_dbm)
    #if guard_str != "true":
      #guard_str = guard_str[1:-1]
      #guard_str = guard_str.replace("&","&amp;")
      #guard_str = guard_str.replace("<","&lt;")
    #else:
      #guard_str = ""
    #if self.guard_bdd:
      #guard_bdd_str = self.template.bdd_to_str(self.guard_bdd)
      #if guard_bdd_str != "true":
	#guard_bdd_str = guard_bdd_str.replace("~","not ")
	#if guard_str:
	  #guard_str += " and (" + guard_bdd_str + ")"
	#else:
	  #guard_str = "(" + guard_bdd_str + ")"
    #if self.guard_var:
      #if guard_str:
	#guard_str += " and (" + self.guard_var + ")"
      #else:
	#guard_str = "(" + self.guard_var + ")"
	
    if self.guard:
      guard_xml = self.guard
      guard_xml = guard_xml.replace("&","&amp;")
      guard_xml = guard_xml.replace("<","&lt;")
      guard_xml = guard_xml.replace(">","&gt;")
      ret += "<label kind=\"guard\">" + guard_xml + "</label>"
      
    if self.synchronisation:
      ret += "<label kind=\"synchronisation\">" + self.synchronisation[0] + self.synchronisation[1] + "</label>"
    
    #update_str = ""
    #for c in self.update_clock:
      #if update_str:
	#update_str += ", " + c.name + "= 0"
      #else:
      	#update_str = c.name + "= 0"
    #for (vname,expr) in self.update_var:
      #if update_str:
	#update_str += ", " + vname + "=" + expr 
      #else:
    	#update_str = vname + "=" + expr
    if self.update:
      ret += "<label kind=\"assignment\">" + self.update + "</label>"
    
    ret += "</transition>"
    return ret

  def has_next(self,state):
    """ Check if the guard of the transition is satisfied by the state"""
    if state.location != self.source:
      return False
    new_state = pysymbolicstates.DiscreteState(self.target,state.variables)
    if self.guard_var:
      print self.guard_var
      val_guard = new_state.evaluate_expression(self.guard_var)
      assert(isinstance(val_guard,bool))
      return val_guard
    else:
      return True

  def next_state(self,state):
    """ Computes the next discrete state when firing the transition for 'state' """
    if state.location != self.source:
      return None
    new_state = pysymbolicstates.DiscreteState(self.target,state.variables)
    if self.guard_var:
      if self.has_next(state):
	for (vname,sign,expr) in self.update_var:
	  val = new_state.evaluate_expression(expr)
	  var = None
	  if self.template.variables.has_key(vname):
	    var = self.template.variables[vname]
	  else:
	    logger.error("Unknown variable in assignment: " + vname + sign + "=" + expr)
	  if sign == "+":
	    val = new_state.variables[var.id] + val
	  else:
	    val = new_state.variables[var.id] - val
	  if val > var.max_val or val < var.min_val:
	    logger.error("Variable range exceeded after assignment: " + vname + sign + "=" + expr)
	  new_variables = list(new_state.variables)
	  new_variables[var.id] = val
	  new_state = pysymbolicstates.DiscreteState(new_state.location,new_variables)
      else:
	new_state = None
    return new_state
      	  
  def previous_state(self,state):
    """ Computes the previous discrete state that allows to reach 'state' through the transition"""
    if state.location != self.target:
      return None
    new_state = pysymbolicstates.DiscreteState(self.source,state.variables)
    if self.guard_var:
      for (vname,sign,expr) in reversed(self.update_var):
	val = new_state.evaluate_expression(expr)
	var = None
	if self.template.variables.has_key(vname):
	  var = self.template.variables[vname]
	else:
	  logger.error("Unknown variable in assignment: " + vname + sign + "=" + expr)
	if sign == "+":
	  val = state.variables[var.id] - val
	else:
	  val = state.variables[var.id] + val
	if val > var.max_val or val < var.min_val:
	  logger.error("Variable range exceeded after reversing assignment: " + vname + sign + "=" + expr)
	new_variables = list(state.variables)
	new_variables[var.id] = val
	new_state = pysymbolicstates.DiscreteState(new_state.location,new_variables)
      if not self.has_next(new_state):
	return None
    return new_state
    
  def parse_guard(self):
    """ Split the guard string into three guards (for clocks, features, and variables) and creates their respective domain (DBM or BDD)"""
    self.guard_dbm = self.template.context.getTautologyFederation()
    self.guard_bdd = None
    self.guard_var = ""
    constraints = split_constraints(self.guard)
    
    if len(self.template.features) > 0:
      self.guard_bdd = self.template.mgr.One()
    for cons in constraints:
      cons = cons.strip()
      if cons:
	dbm = self.template.parse_dbm(cons)
	if dbm:
	  self.guard_dbm &= dbm
	  continue
	if len(self.template.features) > 0:
	  bdd = self.template.parse_bdd(cons)
	  if isinstance(bdd, pycudd.DdNode):
	    self.guard_bdd &= bdd
	    continue
	if self.guard_var:
	  self.guard_var += " and " + cons
	else:
	  self.guard_var = cons

  def parse_update(self):
    """ Split the update string into updates for clocks, updates for feature (that are directly added to the BDD of the guard) and updates for variables """
    self.update_clock = []
    self.update_var = []
    if self.update:
      for u in self.update.split(','):
	u = u.strip()
	#print "update: " + u
	# test if clock reset
	r = re.search('^\s*(\w+)\s*=\s*(\d+)', u)
	if r:
	  varname = r.group(1)
	  if self.template.clocks.has_key(varname):
	    clock = self.template.clocks[varname]
	    value = int(r.group(2))
	    if value != 0:
	      logger.error("Clock reset to non-zero value")
	    else:
	      self.update_clock.append(clock)
	      self.template.updateBound(clock,value)
	    continue
	r = re.search('^\s*(\w+)\s*=(.*)', u)
	if r:
	  # test if adaptive feature update
	  varname = r.group(1)
	  update_expr = r.group(2)
	  if self.template.features.has_key(varname):
	    cons = "(" + varname + " and " + update_expr + ") or ( not " + varname + " and not " + update_expr + ")"
	    bdd = self.template.parse_bdd(cons)
	    if not bdd is None:
	      self.guard_bdd &= bdd
	    else:
	      logger.error("Cannot parse update expression for feature: "+varname)
	  else:  # we assume that the update is for a variable
	    logger.error("Assignement is not allowed exccept for clocks and feature variables.")
	    #varname = self.template.parse_variables_expression(varname)
	    #update_expr =  self.template.parse_variables_expression(update_expr)
	    #self.update_var += (varname,update_expr)
	  continue
	r = re.search('^\s*(\w+)\s*(\+|\-)=(.*)', u)
	if r:
	  varname = r.group(1)
	  update_expr = r.group(3)
	  self.update_var.append((varname,r.group(2),update_expr))
	  continue
	r = re.search('^\s*(\w+)\+\+', u)
	if r:
	  varname = r.group(1)
	  update_expr = "1"
	  self.update_var.append((varname,"+",update_expr))
	  continue
	r = re.search('^\s*(\w+)\-\-', u)
	if r:
	  varname = r.group(1)
	  update_expr = "1"
	  self.update_var.append((varname,"-",update_expr))
	  continue
	logger.error('Syntax error in assignment \"' + u + "\" of transition from \"" + self.source.name + "\" to \"" + self.target.name + "\"")
	  
  def get_guard_bdd(self):
    """ Return the bdd that only conditions the needed features (not update of the adaptive features)
	Both environment and system adaptive features are abstracted, in case the guard was not written properly
    """
    #print "Guard bdd"
    ret = self.guard_bdd
    for f in self.template.fb + self.template.ff:
      ret = ret.ExistAbstract(self.template.features[f])
    #print self.template.bdd_to_str(ret)
    return ret
  
  def get_channel(self):
    return self.synchronisation[0]
  
  def same_channel(self,trans):
    return self.synchronisation[0] == trans.synchronisation[0]
  
  
class Variable:
  def __init__(self,name,init_val,min_val,max_val):
    self.id = 0
    self.name = name
    self.init_val = init_val
    self.min_val = min_val
    self.max_val = max_val
    
  def get_values(self):
    return range(self.min_val,self.max_val+1) 
  