"""PyECDAR, a open source software for timed systems.
Copyright (C) 2013 INRIA Rennes

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA."""

import logging

from pyparsing import Word, Combine, Optional, Forward, Group, Literal, Keyword, ZeroOrMore, OneOrMore, Suppress, alphas, alphanums, nums, stringEnd

logging.basicConfig()
logger = logging.getLogger('PYTADACTL')
    
class TAdaCTLFormula:
    def __init__(self,goal_template,objective):
      self.goal_template = goal_template
      self.timed = False
      if type(objective) == str:
	self.formula = self.parse_formula(objective)
	self.formula_str = objective
      else:
	self.formula = objective.formula
	self.formula_str = objective.formula_str
	self.timed = objective.timed

    def parse_formula(self,formula_str):
	#tokens
	
	comma 		= Literal(",")
	dot 		= Literal(".")
	lparen 		= Literal("(")
	rparen 		= Literal(")")
	lbrack 		= Literal("[")
	rbrack 		= Literal("]")
	equals 		= Literal("==")
	nequals 	= Literal("!=")
	lessequals 	= Literal("<=")
	less 		= Literal("<")
	greater 	= Literal(">")
	greaterequals	= Literal(">=")
	minus		= Literal("-")
	always		= Literal("A")
	exists		= Literal("E")
	until		= Literal("U")
	weak		= Literal("W")
	addop		= Literal("+")
	subop		= Literal("-")
	mulop		= Literal("*")
	divop		= Literal("/")
	modop		= Literal("%")
	orop 		= Keyword("or")
	andop      	= Keyword("and")
	notop       	= Keyword("not")
	trueval       	= Keyword("true")
	falseval       	= Keyword("false")	
	fa		= Keyword("Fa")
	fs		= Keyword("Fs")
	fe		= Keyword("Fe")
	inf 		= Keyword("inf")	 
	
	identifier = Word( alphas, alphanums + "_" )
	number = Word(nums)
	integer = Combine( Optional(minus) + number )
	
	# grammar
	
	#guard_identifier = Combine(identifier + Optional(dot + identifier), adjacent=True)
	
	guard_expr = Forward()
	atomicExpr = Combine(Optional(minus) + identifier | integer | Suppress(lparen) + guard_expr + Suppress(rparen), adjacent=False)
	
	mul_expr = Forward()
	mul_expr << Combine(atomicExpr + Optional((mulop | divop | modop)  + mul_expr), adjacent=False)	

	guard_expr << Combine(mul_expr + Optional((addop | subop)  + guard_expr), adjacent=False)
	
	interval = Group(Suppress(lbrack) + guard_expr + Suppress(comma) + guard_expr + Suppress(rbrack)).setParseAction(lambda t : self.setTimed(True))
	
	orExpression = Forward()
	
	atomicExpression = identifier | (lparen + orExpression + rparen)
	
	notExpression = (Optional(notop) + atomicExpression)

	andExpression = Forward()
	andExpression << (notExpression + Optional(andop + andExpression))
	
	orExpression << andExpression + Optional(orop + orExpression)
	
	featureFormula = Forward()

	compop = equals | nequals | lessequals | less | greaterequals | greater
	
	guard = Combine(guard_expr + compop + guard_expr, adjacent=False).setParseAction(lambda t : self.parse_dbm(t[0]))
			
	loc = Combine(identifier).setParseAction(lambda t : self.parse_location(t[0]))
	
	atomicFormula = Group(trueval.setParseAction(lambda t : self.parse_true_false(True)) | \
			falseval.setParseAction(lambda t : self.parse_true_false(False)) | \
			guard | \
			loc | \
			(Suppress(lparen) + featureFormula + Suppress(rparen))
			)
			
	notFormula = Group(Optional(notop) + atomicFormula)
	
	andFormula = Forward()
	andFormula << Group(notFormula + Optional(andop + andFormula))
	
	orFormula = Forward()
	orFormula << Group(andFormula + Optional(orop + orFormula))
	
	pathFormula = ((Suppress(lparen) + featureFormula + Suppress(rparen) + until + Optional(interval) + Suppress(lparen) + featureFormula + Suppress(rparen)) | \
		       (Suppress(lparen) + featureFormula + Suppress(rparen) + weak + Optional(interval) + Suppress(lparen) + featureFormula + Suppress(rparen)) | \
		       (less + greater + Optional(interval) + Suppress(lparen) + featureFormula + Suppress(rparen)).setParseAction(lambda t : [self.parse_true_false(True), "U"] + t[2:]) | \
		       (lbrack + rbrack + Suppress(lparen) + featureFormula + Suppress(rparen)).setParseAction(lambda t : [t[2], "W", self.parse_true_false(False)]))
	# | Combine(delta + compop) + integer + featureFormula | \  
	
	stateFormula = (Group(always + pathFormula) | Group(exists + pathFormula) | orFormula)
		
	featureFormula << Group(Optional((Suppress(lbrack) + orExpression + Suppress(rbrack)).setParseAction(lambda t : self.parse_bdd(t))) + stateFormula)
	
	return (featureFormula+stringEnd).parseString(formula_str)
	
	
    #def test(self,string,obj):
      #print string + " " + str(obj) + "\n"
      
      
    def setTimed(self,val):
      #print "setting timed"
      self.timed = val
      
    def parse_bdd(self,constraints):
      #print "parse bdd " + " ".join(constraints) + "\n"
      return ["<bdd>"," ".join(constraints)]
      
    def parse_dbm(self,constraints):
      #print "parse dbm " + constraints + "\n"
      return ["<guard>",constraints]
      
    def parse_location(self,loc):
      #print "parse location " + loc + "\n"
      for l in self.goal_template.locations:
	if str(l.name) == str(loc):
	  return ["<loc>",l]
      logger.error("Unknown location \"" + loc + "\" in formula")
      return []

    def parse_true_false(self,val):
      return ["<bool>",val]
