"""PyECDAR, a open source software for timed systems.
Copyright (C) 2013 INRIA Rennes

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA."""

import sys
import logging
import copy

import pysymbolicstates
import pytadactl

logging.basicConfig()
logger = logging.getLogger('TIGA')


""" Timed game algorithms. Two kind of algorithms are implemented:
- game algorithms to check T-AdaCTL formula
- safety game algorithm
"""

class TimedGame:
    """ Implements algorithms to check T-AdaCTL formulae on a Feature Timed Game Automata (FTGA) """
    def __init__(self,goal_template,objective):
        self.goal_template = goal_template	# the FTGA
        if type(objective) == pytadactl.TAdaCTLFormula:
	  self.objective = objective
        else:
	  self.objective = pytadactl.TAdaCTLFormula(self.goal_template,objective)
        if self.objective.timed:		# if the formula has timed operators, the template must be parse with an additional clock
	  self.goal_template.parse_template(True)
	else:  
	  self.goal_template.parse_template()
        
        self.context = self.goal_template.context
        self.mgr = self.goal_template.mgr
        self.results = None			# the set of winning states computed after the game has been solved
	self.initial_discrete_state = self.goal_template.get_initial_discrete_state()	# initial state

    def solve(self):
	""" Solve the timed game and prints reduced results"""
	self.results = self._solve(self.objective.formula)
	self.print_results()
    
    def get_winning_features(self):
	""" Returns the winning features that allows to verify the formula. Returns a BDD object from the PyCUDD library"""
	bdd = self.mgr.Zero()
	if self.results.has_key(self.initial_discrete_state):
	  for sym_state in self.results[self.initial_discrete_state]:
	    if sym_state.zone.hasZero():
	      bdd |= sym_state.bdd
	return bdd
      
    def get_deadlock_state(self,location,controllable):
	""" Get the states beyond the invariant (so in deadlock) because one player didn't do the action he was forced to do. """
	bound_invariant = location.invariant_dbm & ~location.invariant_dbm.tightenUp()
	deadlocks = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
        for t in location.forward:
	  if t.controllable == controllable:
	    guard_inv_bdd = t.get_guard_bdd() & location.invariant_bdd
	    guard_inv_dbm = t.guard_dbm & bound_invariant
	    if not guard_inv_bdd.Empty() and not guard_inv_dbm.isEmpty():
	      deadlocks.append(pysymbolicstates.SymbolicState(guard_inv_bdd,guard_inv_dbm.up()-guard_inv_dbm))
	  #print "BDD non deadlock:"
	  #print self.goal_template.bdd_to_str(bdd_not_deadlock)
	return deadlocks

    def _solve(self,formula):
	""" Internal solve function that recursively solve each nested formula """
      	if type(formula) == dict:
	  assert(len(formula) == len(self.goal_template.locations))
	  return formula
	
      	if len(formula) == 0:
	  logger.error("No formula provided.")
	  return {}
	
	if len(formula) == 1:
	  return self._solve(formula[0])
  
	if len(formula) == 2:
	  if formula[0] == "<bool>":
	    return self.solve_true_false(formula[1])
	  if formula[0] == "<loc>":
	    return self.solve_location(formula[1])
	  if formula[0] == "<guard>":
	    return self.solve_guard(formula[1])
	  if formula[0] == "not":
	    formula[1] = self._solve(formula[1])
	    return self.solve_not(formula[1])

	if len(formula) == 3:
	  if formula[0] == "<bdd>":
	    formula[2] = self._solve(formula[2])
	    return self.solve_feature(formula[1],formula[2])
	  else:
	    formula[0] = self._solve(formula[0])
	    formula[2] = self._solve(formula[2])  
	    if type(formula[1]) == str and formula[1] == "and":
	      return self.solve_and(formula[0],formula[2])
	    if type(formula[1]) == str and  formula[1] == "or":
	      return self.solve_or(formula[0],formula[2])
	    
	if len(formula) == 4:
  	  formula[1] = self._solve(formula[1])
	  formula[3] = self._solve(formula[3])  
	  if type(formula[0]) == str and formula[0] == "A":    
	    if type(formula[2]) == str and formula[2] == "U":
	      return self.solve_A_until(formula[1],formula[3])
	    if type(formula[2]) == str and formula[2] == "W":
	      return self.solve_A_weakuntil(formula[1],formula[3])
	  if type(formula[0]) == str and formula[0] == "E":    
	    if type(formula[2]) == str and formula[2] == "U":
	      return self.solve_E_until(formula[1],formula[3])
	    if type(formula[2]) == str and formula[2] == "W":
	      return self.solve_E_weakuntil(formula[1],formula[3])
	    
	if len(formula) == 5:
  	  formula[1] = self._solve(formula[1])
	  formula[4] = self._solve(formula[4])
	  minval = formula[3][0]
	  maxval = formula[3][1]
	  if type(formula[0]) == str and formula[0] == "A":    
	    if type(formula[2]) == str and formula[2] == "U":
	      return self.solve_A_until(formula[1],formula[4],minval,maxval)
	  if type(formula[0]) == str and formula[0] == "E":    
	    if type(formula[0]) == str and formula[2] == "U":
	      return self.solve_E_until(formula[1],formula[4],minval,maxval)
	    
	logger.error("Unknown formula: " + str(formula))
	return {}
      
    def init_tadactl_clock(self,minval,maxval,psi):
	""" Initialize the additional clock for timed formula.
	    The clock is initialized to zero in all the final states psi
	"""
      	assert(minval != "0" or maxval != "inf")
	tadactl_clock = self.context["tadactl_clock"]
	minval = int(minval)
	tadactl_dbm = tadactl_clock >= minval
	if maxval != "inf":
	  maxval = int(maxval)
	  self.goal_template.updateBound(tadactl_clock,maxval)
	  tadactl_dbm &= tadactl_clock <= maxval
	for state in psi:
	  psi[state].conjunction_in_place(pysymbolicstates.SymbolicState(self.mgr.One(),tadactl_dbm))
	return tadactl_clock
    
    def solve_true_false(self,val):
	""" Solve a atomic formula with either a true or a false proposition
	    - if val is true returns all the symbolic states (since they are all winning)
	    - otherwise returns an empty set 
	"""
	#print "*** Solving " + str(val) + " ***"
	result = {}
	if val:
	  for state in self.goal_template.get_list_discrete_states():
	    sym_states = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
	    sym_states.append(pysymbolicstates.SymbolicState(copy.copy(state.location.invariant_bdd),state.location.invariant_dbm.copy()))
	    result[state] = sym_states
	return result

    def solve_location(self,loc):
      	""" Solve a atomic formula with a location name for proposition
	    Returns winning states if they match the location.
      	"""
	#print "*** Solving location " + loc.name + " ***"
	result = {}
	for state in self.goal_template.get_list_discrete_states():
	  if state.location == loc:
	    sym_states = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
	    sym_states.append(pysymbolicstates.SymbolicState(copy.copy(loc.invariant_bdd),loc.invariant_dbm.copy()))
	    result[state] = sym_states
	  
	#for l in self.goal_template.locations:
	  #result[l] =  pysymbolicstates.SetOfSymbolicStates(self.goal_template)
	  #if l == loc:
	    #result[l].append(pysymbolicstates.SymbolicState(copy.copy(l.invariant_bdd),l.invariant_dbm.copy()))
	  #else:
	    #result[l].append(pysymbolicstates.SymbolicState(copy.copy(l.invariant_bdd),self.goal_template.context.getEmptyFederation()))
	return result
    
    def solve_guard(self,constraints):
        """ Solve a atomic formula with a clock constraint for proposition
	    Returns all the winning states such that this constraint is satisfied.
      	"""
	#print "*** Solving guard " + constraint + " ***"
        result = {}
        zone = self.goal_template.parse_dbm(constraints)
        if zone:
	  for state in self.goal_template.get_list_discrete_states():
	    sym_states = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
	    sym_states.append(pysymbolicstates.SymbolicState(copy.copy(state.location.invariant_bdd),zone & state.location.invariant_dbm))
	    result[state] = sym_states
        else:
	  for state in self.goal_template.get_list_discrete_states():
	    val = state.evaluate_expression(constraints,self.goal_template)
	    assert isinstance(val,bool)
	    if val:
	      sym_states = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
	      sym_states.append(pysymbolicstates.SymbolicState(copy.copy(state.location.invariant_bdd),state.location.invariant_dbm.copy()))
	      result[state] = sym_states
	return result
    
    def solve_not(self,phi):
        """ Solve a 'not' formula.
	    Returns the contrary of the set of winning states phi.
      	"""
	#print "*** Solving not ***"
	result = {}
	for state in self.goal_template.get_list_discrete_states():
	  if phi.has_key(state):
	    sym_states = ~phi[state]
	    sym_states.conjunction_in_place(pysymbolicstates.SymbolicState(state.location.invariant_bdd,state.location.invariant_dbm))
	    if sym_states:
	      result[state] = sym_states
	  else:
	    sym_states = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
	    sym_states.append(pysymbolicstates.SymbolicState(copy.copy(state.location.invariant_bdd),state.location.invariant_dbm.copy()))
	    result[state] = sym_states
	return result
    
    def solve_feature(self,constraints,phi):
        """ Solve a feature formula with a feature expression 'constraints'.
	    Returns the winning states that either do not satisfy 'constraints' or are winning in phi.
      	"""
	#print "*** Solving feature ***"
	bdd = self.goal_template.parse_bdd(constraints)
	if bdd is None:
	  logger.error('Cannot parse boolean expression \'' + constraints + '\' in formula.')
	  return {}
      	result = {}
      	for state in self.goal_template.get_list_discrete_states():
	  if phi.has_key(state):
	    sym_states = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
	    for sym_state in phi[state]:
	      sym_states.append(pysymbolicstates.SymbolicState(bdd & sym_state.bdd, sym_state.zone))
	    sym_states.append(pysymbolicstates.SymbolicState(~bdd & state.location.invariant_bdd, state.location.invariant_dbm.copy()))
	    result[state] = sym_states
	  else:
    	    sym_states = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
	    sym_states.append(pysymbolicstates.SymbolicState(~bdd & state.location.invariant_bdd, state.location.invariant_dbm.copy()))
	    result[state] = sym_states
	return result
	  
    def solve_or(self,phi,psi):
	""" Solve an 'or' formula.
	    Returns the disjunction of the set of states phi and psi.
      	"""
	#print "*** Solving or ***"
	result = {}
	for state in phi:
	  if psi.has_key(state):
	    result[state] = phi[state] | psi[state]
	  else:
	    result[state] = phi[state].copy()
	for state in psi:
	  if not phi.has_key(state):
	    result[state] = psi[state].copy()
	return result
    
    def solve_and(self,phi,psi):
	""" Solve an 'and' formula.
	    Returns the conjunction of the set of states phi and psi.
      	"""
	#print "*** Solving and ***"
      	result = {}
      	for state in phi:
	  if psi.has_key(state):
	    result[state] = phi[state] & psi[state]
	return result

    def solve_A_until(self,phi,psi,minval="0",maxval="inf"):      	
	#print "*** Solving A( )U( ) ***"
	tadactl_clock = None
	if minval != "0" or maxval != "inf":
	  tadactl_clock = self.init_tadactl_clock(minval,maxval,psi)
	  
	winning = psi
	waiting = winning.keys()
	new_states = True
        while waiting:
	  state = waiting.pop(0)
	  loc = state.location
	  loc_inv = pysymbolicstates.SymbolicState(loc.invariant_bdd,loc.invariant_dbm)
          
          good = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
          if winning.has_key(state):
	    good |= winning[state]
	  good |= self.get_deadlock_state(loc,False) & ~self.get_deadlock_state(loc,True) # the system wins if the environnement must do a transition to avoid a deadlock and the system can't
	  for t in loc.forward:
	    if t.controllable:
	      next_state = t.next_state(state)
	      if next_state and winning.has_key(next_state):
		good |= winning[next_state].pred(t)
  	  #print "good states in " + state.to_str() + ": " + good.to_str()

	  bad = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
	  bad.append(loc_inv.copy())
	  if phi.has_key(state):
	    bad &= ~(phi[state])
	  if winning.has_key(state):
	    bad &= ~(winning[state])
	  for t in loc.forward:
	    if not t.controllable:
      	      next_state = t.next_state(state)
	      if next_state:
		if winning.has_key(next_state):
		  bad |= (~winning[next_state]).pred(t)
		else:
  		  next_loc_inv = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
		  next_loc_inv += [pysymbolicstates.SymbolicState(t.target.invariant_bdd,t.target.invariant_dbm)]
		  bad |= next_loc_inv.pred(t)
    	  #print "bad states in " + state.to_str() + ": " + bad.to_str()

	  new_winning = good.predt(loc,bad)
  	  new_winning.conjunction_in_place(loc_inv)
    	  #print "new winning states in " + state.to_str() + ": " + new_winning.to_str()
  	  
	  if winning.has_key(state):
	    for sym_state in new_winning:
	      if winning[state].append(sym_state):
		new_states = True
	  elif new_winning:
	    winning[state] = new_winning
	    new_states = True
    	  #print "winning states in " + state.to_str() + ": " + winning[state].to_str()
	  if new_states:
	    for t in loc.backward:
	      previous_state = t.previous_state(state)
	      if previous_state and previous_state not in waiting:
		waiting.append(previous_state)
	    new_states = False
	
	if tadactl_clock:
	  for state in winning:
	    winning[state].conjunction_in_place(pysymbolicstates.SymbolicState(self.mgr.One(),tadactl_clock == 0))
	    winning[state].freeClockInPlace(tadactl_clock)
	return winning
	      
    def solve_A_weakuntil(self,phi,psi,minval="0",maxval="inf"):
	#print "*** Solving A( )W( ) ***"
	tadactl_clock = None
	if minval != "0" or maxval != "inf":
	  tadactl_clock = self.init_tadactl_clock(minval,maxval,psi)
	
	losing = {}
	waiting = []
	for state in self.goal_template.get_list_discrete_states():
	  loc_inv = pysymbolicstates.SymbolicState(state.location.invariant_bdd,state.location.invariant_dbm)
	  bad = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
	  bad.append(loc_inv.copy())
	  if phi.has_key(state):
	    bad &= ~(phi[state])
	  if psi.has_key(state):
	    bad &= ~(psi[state])
	  if bad:
	    losing[state] = bad
	    waiting.append(state)
	
	new_states = True
        while waiting:
	  state = waiting.pop(0)
	  loc = state.location
	  loc_inv = pysymbolicstates.SymbolicState(loc.invariant_bdd,loc.invariant_dbm)
          
          bad = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
	  if losing.has_key(state):
	    bad |= losing[state]
      	  #print "bad states in " + state.to_str() + ": " + bad.to_str()
          bad |= self.get_deadlock_state(loc,True) # the environnement wins if the system can do a transition to avoid a deadlock
	  for t in loc.forward:
	    if not t.controllable:
	      next_state = t.next_state(state)
	      if next_state and losing.has_key(next_state):
		bad |= losing[next_state].pred(t)
    	  #print "bad states in " + state.to_str() + ": " + bad.to_str()

          good = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
          if psi.has_key(state):
	    good |= psi[state]
	  for t in loc.forward:
	    if t.controllable:
      	      next_state = t.next_state(state)
	      if next_state:
		if losing.has_key(next_state):
		  good |= (~losing[next_state]).pred(t)
		else:
		  next_loc_inv = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
		  next_loc_inv += [pysymbolicstates.SymbolicState(t.target.invariant_bdd,t.target.invariant_dbm)]
		  good |= next_loc_inv.pred(t)
  	  #print "good states in " + state.to_str() + ": " + good.to_str()

	  new_losing = bad.predt(loc,good-bad)
  	  new_losing.conjunction_in_place(loc_inv)
    	  #print "new losing states in " + state.to_str() + ": " + new_losing.to_str()

	  if losing.has_key(state):
	    for sym_state in new_losing:
	      if losing[state].append(sym_state):
		new_states = True
	  elif new_losing:
	    losing[state] = new_losing
	    new_states = True
    	  #print "losing states in " + state.to_str() + ": " + losing[state].to_str()
	  if new_states:
	    for t in loc.backward:
      	      previous_state = t.previous_state(state)
	      if previous_state and previous_state not in waiting:
		waiting.append(previous_state)
	    new_states = False
	       
	winning = {}
	for state in self.goal_template.get_list_discrete_states():
	  loc_inv = pysymbolicstates.SymbolicState(state.location.invariant_bdd,state.location.invariant_dbm)
	  good = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
	  good.append(loc_inv.copy())
	  if losing.has_key(state):
	    good &= (~losing[state])
	  if tadactl_clock:
	    good.conjunction_in_place(pysymbolicstates.SymbolicState(self.mgr.One(),tadactl_clock == 0))
	    good.freeClockInPlace(tadactl_clock)
	  if good:
	    winning[state] = good
	return winning

    def solve_E_until(self,phi,psi,minval="0",maxval="inf"):
	#print "*** Solving E( )U( ) ***"
	tadactl_clock = None
	if minval != "0" or maxval != "inf":
	  tadactl_clock = self.init_tadactl_clock(minval,maxval,psi)
	
	winning = psi
	waiting = winning.keys()
	new_states = True
        while waiting:
	  state = waiting.pop(0)
	  loc = state.location
	  loc_inv = pysymbolicstates.SymbolicState(loc.invariant_bdd,loc.invariant_dbm)
          
          good = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
          if winning.has_key(state):
	    good |= winning[state]
	  for t in loc.forward:
	    next_state = t.next_state(state)
	    if next_state and winning.has_key(next_state):
	      good |= winning[next_state].pred(t)
  	  #print "good states in " + state.to_str() + ": " + good.to_str()

	  bad = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
	  bad.append(loc_inv.copy())
	  if phi.has_key(state):
	    bad &= ~(phi[state])
	  if winning.has_key(state):
	    bad &= ~(winning[state])
    	  #print "bad states in " + state.to_str() + ": " + bad.to_str()

	  new_winning = good.predt(loc,bad)
  	  new_winning.conjunction_in_place(loc_inv)
    	  #print "new winning states in " + state.to_str() + ": " + new_winning.to_str()

	  if winning.has_key(state):
	    for sym_state in new_winning:
	      if winning[state].append(sym_state):
		new_states = True
	  elif new_winning:
	    winning[state] = new_winning
	    new_states = True
    	  #print "winning states in " + state.to_str() + ": " + winning[state].to_str()
	  if new_states:
	    for t in loc.backward:
	      previous_state = t.previous_state(state)
	      if previous_state and previous_state not in waiting:
		waiting.append(previous_state)
	    new_states = False
	
	if tadactl_clock:
	  for state in winning:
	    winning[state].conjunction_in_place(pysymbolicstates.SymbolicState(self.mgr.One(),tadactl_clock == 0))
	    winning[state].freeClockInPlace(tadactl_clock)
	return winning
	      
    def solve_E_weakuntil(self,phi,psi):
	#print "*** Solving E( )W( ) ***"
	tadactl_clock = None
	if minval != "0" or maxval != "inf":
	  tadactl_clock = self.init_tadactl_clock(minval,maxval,psi)
	
	losing = {}
	waiting = []
	for state in self.goal_template.get_list_discrete_states():
	  loc_inv = pysymbolicstates.SymbolicState(state.location.invariant_bdd,state.location.invariant_dbm)
	  bad = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
	  bad.append(loc_inv.copy())
	  if phi.has_key(state):
	    bad &= ~(phi[state])
	  if psi.has_key(state):
	    bad &= ~(psi[state])
	  if bad:
	    losing[state] = bad
	    waiting.append(state)
	
	new_states = True
	while waiting:
	  state = waiting.pop(0)
	  loc = state.location
	  loc_inv = pysymbolicstates.SymbolicState(loc.invariant_bdd,loc.invariant_dbm)
          
          bad = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
	  if losing.has_key(state):
	    bad |= losing[state]
          bad |= self.get_deadlock_state(loc,False) | self.get_deadlock_state(loc,True) # the system loses if either the environnement or the system must do a transition to avoid a deadlock
    	  #print "bad states in " + state.to_str() + ": " + bad.to_str()

          good = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
          if psi.has_key(state):
	    good |= psi[state]
	  for t in loc.forward:
	    next_state = t.next_state(state)
	    if next_state:
	      if losing.has_key(next_state):
		good |= (~losing[next_state]).pred(t)
	      else:
		next_loc_inv = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
		next_loc_inv += [pysymbolicstates.SymbolicState(t.target.invariant_bdd,t.target.invariant_dbm)]
		good |= next_loc_inv.pred(t)
  	  #print "good states in " + state.to_str() + ": " + good.to_str()

	  new_losing = bad.predt(loc,good-bad)
  	  new_losing.conjunction_in_place(loc_inv)
    	  #print "new losing states in " + state.to_str() + ": " + new_losing.to_str()

	  if losing.has_key(state):
	    for sym_state in new_losing:
	      if losing[state].append(sym_state):
		new_states = True
	  elif new_losing:
	    losing[state] = new_losing
	    new_states = True
	  #print "losing states in " + state.to_str() + ": " + losing[state].to_str()
	  if new_states:
	    for t in loc.backward:
      	      previous_state = t.previous_state(state)
	      if previous_state and previous_state not in waiting:
		waiting.append(previous_state)
	    new_states = False
	       
	winning = {}
	for state in self.goal_template.get_list_discrete_states():
	  loc_inv = pysymbolicstates.SymbolicState(state.location.invariant_bdd,state.location.invariant_dbm)
	  good = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
	  good.append(loc_inv.copy())
	  if losing.has_key(state):
	    good &= (~losing[state])
	  if tadactl_clock:
	    good.conjunction_in_place(pysymbolicstates.SymbolicState(self.mgr.One(),tadactl_clock == 0))
	    good.freeClockInPlace(tadactl_clock)
	  if good:
	    winning[state] = good
	return winning
               
    def print_results(self,all_states=False):
	if self.results == None:
	  logger.error("The game has not been solved yet")
	print "Checking formula \'" + self.objective.formula_str + "\' in template \'" + self.goal_template.name + "\'"
	if all_states:
	  print "=== Winning states ==="
	  for (state,sym_states) in self.results.items():
	    print state + " : "
	    print sym_states
	print "\n=== Winning features ==="
	print "  " + self.goal_template.bdd_to_str(self.get_winning_features())
	
      
class SafetyTimedGame:
    def __init__(self,goal_template,goal_locations_name):
        self.goal_template = goal_template
        self.context = self.goal_template.context
        self.initial_discrete_state = self.goal_template.get_initial_discrete_state()
	self.accepting = {}
       	for l in self.goal_template.locations:
	  self.accepting[l] = (l.name in goal_locations_name)
        self.solved = 0		# the game is initially undetermined
        self.zone = {}		# link a state to a zone (a set of symbolic states, but without features, so a federation)
        self.accepting_zone = {}
        self._forward = {} 	# dictionnary that store for each state the next reachable states
       	self.depend = {}	# link a state to the transitions that may lead to this state
        self.strategy = []	# ordered list of the states refined
	
    def next(self,state,zone):
        ret = []
        for transition in state.location.forward:
	  next_state = transition.next_state(state)
	  if next_state:
	    next_zone = zone.post(transition)
	    if next_zone:
	      ret.append((state,zone,transition,next_state,next_zone))
	      if self.depend.has_key(next_state):
		if not (state,transition) in self.depend[next_state]:
		  self.depend[next_state].append((state,transition))
	      else:
		self.depend[next_state] = [(state,transition)]
	self._forward[state] = ret
	return ret
    
    def nextControllable(self,state):
      	logger.debug("nextControllable")
      	if self._forward.has_key(state):
	  return [e for e in self._forward[state] if e[2].controllable]
	else:
	  return []
    
    def nextUncontrollable(self,state):
      	logger.debug("nextUncontrollable")
      	if self._forward.has_key(state):
	  return [e for e in self._forward[state] if not e[2].controllable]
	else:
	  return []

    def solve(self):
        if self.accepting[self.initial_discrete_state.location]:
	  return False

        self.solved = 0
        self.zone = {}
        self.accepting_zone = {}
        self._forward = {}
        self.depend = {}
        self.strategy = []
        waiting = []
        
	initial_zone = self.context.getZeroFederation()
        initial_zone.upInPlace()
        initial_zone &= self.initial_discrete_state.location.invariant_dbm
        initial_zone.extrapolateMaxBoundsInPlace(self.goal_template.clock_bounds)
	initial_sym_state = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
	initial_sym_state += [pysymbolicstates.SymbolicState(None,initial_zone)]
	self.zone[self.initial_discrete_state] = initial_sym_state
	for e in self.next(self.initial_discrete_state,initial_sym_state):
	  logger.debug("adding to wait next initial: " + str(e[3]) + "#" + str(e[4]))
          waiting.append(e)

	i = 0
        while waiting:
            e = waiting.pop()
            state = e[0]
            loc = e[0].location
            old_zone = e[1]
            transition = e[2]
            next_state = e[3]
            next_loc = e[3].location
            next_zone = e[4]
	    logger.debug(str(i)+':' + str(state) + "#" + str(old_zone) + '-->' + str(next_state) + "#" + str(next_zone) )
            i += 1
            if not self.zone.has_key(next_state) or not next_zone.is_included_in(self.zone[next_state]): # not included
	      logger.debug(str(i)+'Exploring new state:' + str(next_state) + "#" + str(next_zone) + ' [accepting=' + str(self.accepting[next_loc])+ ']' +' (from ' + str(transition) + ')')
	      if self.zone.has_key(next_state):
		self.zone[next_state] |= next_zone
	      else:
		self.zone[next_state] = next_zone
	      if not self.accepting[next_loc]:
		for next_e in self.next(next_state,next_zone):
		  logger.debug("adding to wait next: " + str(next_e[3]) + "#" + str(next_e[4]))
		  if not next_e in waiting: # Can the contrary happen ?
		    waiting.append(next_e)
		  else:
		    logger.debug("already in wait: " + str(next_e[3]) + "#" + str(next_e[4]))
	      else:
		self.accepting_zone[next_state] = self.zone[next_state].copy()
		logger.debug("adding to wait accepting: " + str(next_state))
		waiting.append(e)
            else:                
	      logger.debug(str(i)+'Updating :' + str(state) + "#" + str(old_zone) + '(from ' + str(transition) + ')')
	      bad = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
	      if self.accepting_zone.has_key(state):
		bad |= self.accepting_zone[state]
	      if not self.zone.has_key(state):
		logger.error("Visited state: " * str(state) + " has not reachable zone.")
	      bad |= self.zone[state].up() - self.zone[state]
	      for a in self.nextUncontrollable(state):
		if self.accepting_zone.has_key(a[3]):
 		  #print a[2].source.name + " --> " + a[2].target.name 
		  bad |= (self.accepting_zone[a[3]].pred(a[2]) & old_zone)
	      #print "bad: " + str(bad)
	      good  = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
	      for a in self.nextControllable(state):	
		#print a[2].source.name + " --> " + a[2].target.name 
		if self.accepting_zone.has_key(a[3]):
		  good |= ((a[4] - self.accepting_zone[a[3]]).pred(a[2]) & old_zone)
		else:
		  good |= (a[4].pred(a[2]) & old_zone)
	      #print "good: " + str(good)
	      accepting_ = bad.predt(loc,good-bad)
	      accepting_ &= old_zone
	      #print accepting_

	      if not self.accepting_zone.has_key(state) or not accepting_.is_included_in(self.accepting_zone[state]):
		self.strategy.append(state)
		if self.depend.has_key(state):
		  for (previous_state,t) in self.depend[state]:
		    add = True
		    for w in waiting:
		      if w[0] == previous_state:
			add = False
		    if add:
		      e_ = (previous_state, self.zone[previous_state] , t , state , self.zone[state])
		      logger.debug("Adding to wait dependance: " + str(e_[3]) + "#" + str(e_[4]))
		      waiting.append(e_)
		self.accepting_zone[state] = accepting_
		logger.debug('Update :' + str(state)+ "#" + str(old_zone) + '[accepting=' + str(accepting_)+ ']')
	      if not (state,t) in self.depend[next_state]:
		self.depend[next_state].append((state,t))
	      logger.debug('Accepting zone:' + str(accepting_))
	    if self.accepting_zone.has_key(self.initial_discrete_state) and self.accepting_zone[self.initial_discrete_state].hasZero():
              self.solved = -1
	      break

	logger.info(str(i) + " states visited.")
	if self.solved >=0:
	  self.solved = 1 # game is won by the controller
        return (self.solved > 0)

    def refine_safety(self,delta_value):
      """ Refine the spoiling strategy to extract parametrically the minimum value of delta.
      """
      param_context = self.goal_template.param_context
      initial_domain = param_context.getInitialDomain(0,delta_value)

      if self.solved >= 0 : # the game is winning for the environnement (or not solved)
	return param_context.getEmptyParamFederation()

      accepting_param_zone = {}
      for state in self.goal_template.get_list_discrete_states():
	if self.accepting[state.location]:
	  accepting_param_zone[state] = state.location.invariant_ppl.copy() 
		
      waiting = self.strategy
      i = 0
      while waiting:
	current_state = waiting.pop(0)
	current_loc = current_state.location
	if not self.accepting_zone.has_key(current_state) or self.accepting_zone[current_state].isEmpty():
	  continue
	logger.debug("Refining: " + str(current_state))
	i += 1
        bad = ~current_loc.invariant_ppl
        bad &= initial_domain
        for a in self.nextUncontrollable(current_state):
	  if accepting_param_zone.has_key(a[3]) and not accepting_param_zone[a[3]].isEmpty():
	    bad |= accepting_param_zone[a[3]].pred(a[2].guard_ppl & current_loc.invariant_ppl,a[2].update_ppl)
	good = param_context.getEmptyParamFederation()
	for a in self.nextControllable(current_state):
	  if accepting_param_zone.has_key(a[3]):
	    pred_acc = a[3].location.invariant_ppl - accepting_param_zone[a[3]]
	    good |= pred_acc.pred(a[2].guard_ppl & current_loc.invariant_ppl,a[2].update_ppl) 
	  else:
	    pred_acc = a[3].location.invariant_ppl
	    good |= pred_acc.pred(a[1].guard_ppl & current_loc.invariant_ppl,a[1].update_ppl) 
	good -= bad
	good &= initial_domain
	
	accepting_param_zone[current_state] = bad.predt(current_loc,good)
	accepting_param_zone[current_state] &= current_loc.invariant_ppl
	#accepting_.pairwiseReduce()
	
	logger.debug(str(i) + " states refined;")
	logger.debug("Refined location: " + str(current_state) + " ; accepting domain: " + str(accepting_param_zone[current_state]))
	if current_state == self.initial_discrete_state:
	  refine_values = accepting_param_zone[current_state].hasZeroParamDomain()
	  if not refine_values.isEmpty():
	    logger.info(str(i) + " states refined;")
	    return refine_values.minimizeParam()
      logger.error("No winning value in initial parameter domain.")
      return None	
    
 
            
#if __name__ == '__main__':

    #model = pyuppaal.NTA.from_xml(sys.argv[1])
    
    #dbmpyuppaal.parse_nta_dbm(model)
    
    #goal_template = model.getTemplateByName(sys.argv[2].split('.')[0])
    #goal_location_name = sys.argv[2].split('.')[1]
    
    #dbmpyuppaal.parse_templates_dbm(model,[goal_template])

    #game = SafetyTimedGame(model, goal_template, goal_location_name) 
    #print game.solve()
    #game.printDot('game.dot')

    
